#include "OgreItemBufferQueryManager.h"
#include <OgreLogManager.h>
#include <OgreStringConverter.h>
#include <vector>


namespace Ogre {
//----------------------------------------------------------------------- 
ItemBufferQueryManager::ItemBufferQueryManager(PlatformHierarchyInterface *hierarchyInterface, Viewport *vp):
PlatformQueryManager(hierarchyInterface, vp), mWasInitialised(false)
{
}
//-----------------------------------------------------------------------
bool ItemBufferQueryManager::ShootRay(const Ray &ray, std::vector<Mesh *> *visibleMeshes, bool isGlobalLine)
{
    // run OGRE ray shooting query
    return false;
}
//-----------------------------------------------------------------------
void ItemBufferQueryManager::ComputeCameraVisibility(const Camera &camera,
			    InfoContainer<GtpVisibility::NodeInfo> *visibleNodes,
			    InfoContainer<GtpVisibility::MeshInfo> *visibleGeometry,
			    bool relativeVisibility)
{
	// initialise item buffer if necessary
	InitItemBuffer(visibleNodes, visibleGeometry);

	// we need access to the scene manager and the rendersystem
	PlatformHierarchyInterface *pfHierarchyInterface = 
		dynamic_cast<PlatformHierarchyInterface *>(mHierarchyInterface);

	SceneManager *sm = pfHierarchyInterface->GetSceneManager();

	//-- Render scene as item buffer (e.g., color coded objects)

	// const_cast allowed because camera is not changed in renderScene
	Camera *pCam = const_cast<Camera *>(&camera); 

	// disable overlays, reset them later
	bool overlayEnabled = mViewport->getOverlaysEnabled();
	mViewport->setOverlaysEnabled(false);

	// set item buffer (must be provided by scene manager)
	bool useItemBuffer = true;
	sm->setOption("UseItemBuffer", &useItemBuffer);
	
	// clear background with black
	ColourValue bg = mViewport->getBackgroundColour();
	mViewport->setBackgroundColour(ColourValue(0, 0, 0, 0));

	//pfHierarchyInterface->GetRenderSystem()->clearFrameBuffer(FBT_COLOUR);
	// render item buffer
	pfHierarchyInterface->GetSceneManager()->_renderScene(pCam, mViewport, false);
	
	// reset old overlay status
	mViewport->setOverlaysEnabled(overlayEnabled);
	
	mViewport->setBackgroundColour(bg);

	// don't need item buffer anymore
	useItemBuffer = false;
	sm->setOption("UseItemBuffer", &useItemBuffer);


	//---- collect results
	int dimx = 0;
	int dimy = 0;

	// get frame buffer
	uchar *buf = mViewport->getTarget()->getBufferContents(dimx, dimy);

	// loop through frame buffer & collect visible pixels
	for (int idx = 0; idx < dimy * dimx * 3; idx += 3)
	{
		int id = buf[idx] << 16;
		id += buf[idx + 1] << 8;
		id += buf[idx + 2];

		// if valid id <= add visibility
		if ((id >= 0) && (id < (int)visibleGeometry->size()))
		{
			((*visibleGeometry)[id]).AddVisibility(1, 1);
		}
	/*	else
		{ 
			std::stringstream d; 
			d << "id: " << id << ", buf 1: " << (int)buf[idx] << ", buf 2: " << (int)buf[idx + 1] << ", buf 3: " << (int)buf[idx + 2];
			LogManager::getSingleton().logMessage(d.str());
		}*/
	}

	// reset initialised - flag
	mWasInitialised = false;

	delete [] buf;
}
//-----------------------------------------------------------------------
void ItemBufferQueryManager::InitItemBuffer(InfoContainer<GtpVisibility::NodeInfo> *visibleNodes,
			    InfoContainer<GtpVisibility::MeshInfo> *visibleGeometry)
{
	if (mWasInitialised)
		return;

	mWasInitialised = true;

	SceneManager::EntityIterator it = dynamic_cast<PlatformHierarchyInterface *>(mHierarchyInterface)->
		GetSceneManager()->getEntityIterator();

	// TODO: make this more efficient
	visibleGeometry->clear();

	while (it.hasMoreElements())
	{
		visibleGeometry->push_back(GtpVisibility::MeshInfo(it.getNext(), 0, 0));
	}
}
//-----------------------------------------------------------------------
void ItemBufferQueryManager::ComputeFromPointVisibility(const Vector3 &point,
			       InfoContainer<GtpVisibility::NodeInfo> *visibleNodes,
			       InfoContainer<GtpVisibility::MeshInfo> *visibleGeometry,
			       bool relativeVisibility)
{
	SceneManager *sm = dynamic_cast<PlatformHierarchyInterface *>
		(mHierarchyInterface)->GetSceneManager();

	Camera *cam = sm->createCamera("PointQueryCam");	

	//save old camera
	Camera *savedCam = mViewport->getCamera();
	
	// --- initialise new camera
	mViewport->setCamera(cam);
	cam->setPosition(point);

	cam->setNearClipDistance(savedCam->getNearClipDistance());
	cam->setFarClipDistance(savedCam->getFarClipDistance());

	// set frustum to 45 degrees so all the scene can be captured with 6 shots
	cam->setAspectRatio(1.0);
	cam->setFOVy(Radian(Math::HALF_PI));

	int sign = -1;
	
	// ---- capture visibility from all 6 directions
	for (int i=0; i < 6; i++)	
	{
		sign *= -1;
		
		// Print camera details
      	std::stringstream d;
		d << "Point query camera: " + StringConverter::toString(cam->getDerivedPosition()) + 
			" " + "O: " + StringConverter::toString(cam->getDerivedOrientation());
		LogManager::getSingleton().logMessage(d.str());

		// prevent from initialising array again
		if (i > 0)
			mWasInitialised = true;
		
		// compute camera visibility for this direction
		ComputeCameraVisibility(*cam, visibleNodes, visibleGeometry, relativeVisibility);

		//mViewport->getTarget()->update(); for(int j=0; j<10000000; j++)	printf("wait"); // HACK: DISPLAY CAMERA VIEW

		// --- permute directions
		Vector3 dir(0,0,0);
		dir[i/2] = sign;

		cam->setDirection(dir);
	}
	
	// reset camera
	mViewport->setCamera(savedCam);
}

} // namespace Ogre
